/***
*** init.c  
***
***  General description of this file:
***     Device driver source code for General Standards 16AO12 
***     family Delta-Sigma A/D boards. This file is part of the Linux
***     driver source distribution for this board.
***     
***     This file is not necessary to compile application programs, therefore 
***     should NOT be included in binary only driver distributions.
***
***  Copyrights (c):
***     General Standards Corporation (GSC), 2004-2006
***
***  Author:
***     Evan Hillman (evan@generalstandards.com)
***
***  Support:
***     Primary support for this driver is provided by GSC. 
***
***  Platform (tested on, may work with others):
***     Linux, kernel version 2.4.x, Red Hat distribution, Intel hardware.
***/

//////////////////////////////////////////////////////////////////////////
// set the following flag to trace debug messages in this module.
#ifdef DEBUG
#define TRACE_LOCAL TRUE
#endif

#include <linux/proc_fs.h>
#include "internals.h"

struct board_entry boards_supported[] = {
    {GSC_SUBVENDOR,0x2405/*0x3102*/,"GSC16AO12",0},
    {0,0,"NULL",0},               /* null terminated list.... */
};

/* local module data */
static struct file_operations device_fops = {
open:    device_open,
release: device_close,
read:    device_read,
write:   device_write,
ioctl:   device_ioctl
};

static int device_major = 0;                  /* device major number (dynamically assigned) */
static struct device_board *boards = 0;       /* linked list of all boards found */

__s8    built[128];
int     num_boards;
int     proc_enabled;

/************************************************************************/
/* module initialization: detect card(s) and set up descriptor(s)        */
/************************************************************************/
int init_module(void)
{
    int index = 0;
    unsigned short reg;
    struct device_board *device=NULL, *devicenext;
    struct pci_dev *pdev=NULL;
    struct proc_dir_entry*  proc;
    int i;
    int found;

    sprintf(built, "%s, %s ", __DATE__, __TIME__);
    logmsg(": driver ( version: %s) built %s loading on kernel %s\n",  DRIVER_VERSION, built, UTS_RELEASE);
    logmsg(": Copyright (C)" COPYRIGHT "General Standards Corp. \n");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,74)
    PCI_DEVICE_LOOP(pdev) {
#else
    while((pdev = pci_find_device(PCI_VENDOR_ID_PLX, PCI_ANY_ID, pdev))) {
#endif
        msg(": Checking...vendor: %X device: %X subvendor: %X subsystem: %X\n",pdev->vendor, pdev->device, pdev->subsystem_vendor, pdev->subsystem_device);
        if ((pdev->vendor == PCI_VENDOR_ID_PLX) && (pdev->device == PCI_DEVICE_ID_PLX_9080)) {

            /* determine if this is one of the boards supported. */

            found=0;
            i=0;
            while(boards_supported[i].subsystem_device != 0){
                if ((boards_supported[i].subsystem_device == pdev->subsystem_device) &&
                    (boards_supported[i].subsystem_vendor == pdev->subsystem_vendor)){
                        found=1;
                        //msg(":     found board %s type %d\n",boards_supported[i].name,i);
                    }
                    if (found) break;
                    i++;
            }

            if (found){
                msg(":     installing board %s type %d\n",boards_supported[i].name,i);
                pci_enable_device(pdev);
                pci_read_config_word(pdev, 0x2E, &reg);
                //msg(":     config reg=0x%x\n", reg);
                //msg(":     attaching board #%d\n", index + 1);
                device = (struct device_board *)kmalloc(sizeof(struct device_board), GFP_KERNEL);
                memset(device,0,sizeof(device));
                device ->board_type = i;
                {
                    int ResourceCount = 0;

                    for (i = 0; i < PCI_NUM_BARS; ++i)
                    {
                        // Verify the address is valid
                        if (pci_resource_start(pdev, i) == 0)
                        {
                            continue;
                        }

                        //msg("   Resource %02d\n", ResourceCount);

                        // Increment resource count
                        ResourceCount++;
                        // Get PCI physical address
                        device->PciBar[i].Physical.QuadPart = pci_resource_start(pdev, i);

                        // Determine resource type
                        if (pci_resource_flags(pdev, i) & IORESOURCE_IO)
                        {
                            //msg("     Type     : I/O Port\n");

                            // Make sure flags are cleared properly
                            device->PciBar[i].Physical.QuadPart &= ~(0x3);
                            device->PciBar[i].IsIoMapped         = TRUE;
                        }
                        else
                        {
                            //msg("     Type     : Memory Space\n");

                            // Make sure flags are cleared properly
                            device->PciBar[i].Physical.QuadPart &= ~(0xf);
                            device->PciBar[i].IsIoMapped         = FALSE;
                        }

                        //msg("     Address  : %08x\n",device->PciBar[i].Physical.u.LowPart);

                        // Get the size
                        device->PciBar[i].Size = 4096; /*
                                                       Plx_pci_resource_len(
                                                       device,
                                                       i
                                                       );*/
                    }
                }

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,0)
                device->runtime_addr = (u32 *)ioremap(pdev->base_address[0], 4096);
                device->ioctlwq = NULL;
                device->readwq = NULL;
                device->writewq = NULL;
                device->dmawq = NULL;
                if (device->PciBar[2].IsIoMapped)
                {
                    //device->local_addr = pdev->resource[2];
                    msg(": I/O mapped local registers\n");
                    device->local_addr = base_address[2];
                }
                else
                {
                    device->local_addr = (u32 *)ioremap(pdev->resource[2].start, 4096);
                }
#else
                if (device->PciBar[2].IsIoMapped)
                {
                    device->local_addr = (u32 *)pdev->resource[2].start;
                }
                else
                {
                    device->local_addr = (u32 *)ioremap(pdev->resource[2].start, 4096);
                }
                device->runtime_addr = (u32 *)ioremap(pdev->resource[0].start, pci_resource_len(pdev, 0));
                init_waitqueue_head(&device->ioctlwq);
                init_waitqueue_head(&device->writewq);
                init_waitqueue_head(&device->dmawq);
#endif
                device->pdev = pdev;
                pci_set_master(pdev);
                device->fillBuffer=FALSE;
                device->irqlevel = pdev->irq;
                device->busy = 0;
                device->minor = index;
                device->next = boards;
                msg(":     base_address[0]=0x%lX, base_address[2]=0x%lX\n",pdev->resource[0].start, pdev->resource[2].start);
                //              msg(":     local_addr=%p, runtime_addr=%p irq=%d\n", device->local_addr, device->runtime_addr, device->irqlevel);
                boards = device;
                device->board_index=index;
#ifdef TRACE_LOCAL
                board_type[index]=device->board_type;
#endif
                index++;
                num_boards++;
            }
        }
    }
    if (index == 0) {
        errmsg(": no board found\n");
        return (-ENODEV);
    }
    device_major = register_chrdev(0, GSC_NAME, &device_fops);
    if (device_major < 0) {
        /* OK, could not register -- undo the work we have done above */
        errmsg(": could not register device number\n");
        for (device = boards; device; device = devicenext) {
            devicenext = device->next;
            iounmap(device->local_addr);
            iounmap(device->runtime_addr);
            pci_disable_device(device->pdev);
            kfree(device);
        }
        boards = NULL;
        return (-ENODEV);
    }
    /*
    *   Add /proc file system support.
    */

    if (num_boards)
    {
        remove_proc_entry(GSC_NAME, NULL);  /* Just in case. */
        proc    = create_proc_entry(GSC_NAME, S_IRUGO, NULL);

        if (proc)
        {
            proc_enabled    = 1;
            proc->read_proc = read_proc;
            proc->get_info  = (void*)proc_get_info;
        }
        else
        {
            msg( "<1>%s: create_proc_entry() failure.\n", GSC_NAME);
            cleanup_module();
        }
    }

#ifdef TRACE_LOCAL
    msg(": major=%d\n", device_major);
    {
        int j;
        for (j=0;j<MAX_BOARDS;j++){
            int_other_count[j]=0;
            int_count[j]=0;
            dma_count[j]=0;
            channel_irq[j]=0;
        }
    }
#endif
    return 0;
}

/************************************************************************/
/* cleanup when unloading module                                        */
/************************************************************************/
void cleanup_module(void)
{
    struct device_board *device, *devicenext;
    unregister_chrdev(device_major, GSC_NAME);
    for (device = boards; device; device = devicenext) {
        devicenext = device->next;
        iounmap(device->local_addr);
        iounmap(device->runtime_addr);
        pci_disable_device(device->pdev);
        kfree(device);
    }
    if (proc_enabled)   {
        remove_proc_entry(GSC_NAME, NULL);
        proc_enabled    = 0;
    }

    boards = NULL;
    //msg(": unloaded\n");
}

/************************************************************************/
/* open device                                                          */
/************************************************************************/
int device_open(struct inode *inode, struct file *fp)
{
    struct device_board *device;

    for (device = boards; device; device = device->next) {
        if (MINOR(inode->i_rdev) == device->minor) {
            if (device->busy) {
                errmsg("(%d): board already opened\n", device->minor);
                return (-EBUSY);
            }
            msg("(%d): opening board\n", device->minor);
            
            if (request_irq(device->irqlevel, (void *) device_interrupt, SA_SHIRQ, GSC_NAME, device) < 0) {
                msg("(%d): cannot get interrupt %d\n", device->minor, device->irqlevel);
                device->busy = 0;
                return (-EBUSY);
            }
            if (!(device->dma_data[TX].intermediateBuffer = (u32 *)__get_dma_pages(GFP_KERNEL, DMA_ORDER))) {
                msg("(%d): can not allocate DMA pages\n", device->minor);
                free_irq(device->irqlevel, device);
                device->busy = 0;
                return (-EBUSY);
            }
            
#ifdef TRACE_LOCAL
            //memset(device->intermediateBuffer,0xa5,DMA_ORDER*PAGE_SIZE);
#endif

            device->dma_data[TX].intermediatePhysicalAddr = virt_to_phys((void *)device->dma_data[TX].intermediateBuffer);
            msg("(%d): DMA virt=%p, phys=0x%x size %d\n", device->minor, device->dma_data[TX].intermediateBuffer, device->dma_data[TX].intermediatePhysicalAddr, sizeof(device->dma_data[TX].intermediateBuffer));
            init_timer(&device->watchdog_timer);
            device->timeout_seconds=20;
            device->watchdog_timer.function=timeout_handler;
            device->watchdog_timer.data = (unsigned long) device;
#ifdef DEBUG
            device->nread=0;
            device->last=-1;
#endif
            device->busy = 1;
            device->dma_data[TX].dmaState = DMA_DISABLE;
            device->dma_data[TX].dmastart = FALSE;
            device->dma_data[TX].dmasamples = 0;
            device->dma_data[TX].intermediateSamples=0;
            memset((void *)device->irq_event_pending,FALSE,sizeof(device->irq_event_pending));

            device->error = DEVICE_SUCCESS;
            device->signalno = (-1);
            device->signalev = NO_EVENT;

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,0)
            device->ioctlwq = NULL;
            device->writewq = NULL;
            device->dmawq = NULL;
#else
            init_waitqueue_head(&device->ioctlwq);
            init_waitqueue_head(&device->writewq);
            init_waitqueue_head(&device->dmawq);
#endif          
            device->event_queue_waiting[EVENT_INIT_COMPLETE]=&device->ioctlwq;
            device->event_queue_waiting[EVENT_AUTOCAL_COMPLETE]=&device->ioctlwq;
            device->event_queue_waiting[EVENT_OUT_BUFFER_EMPTY]=&device->writewq;
            device->event_queue_waiting[EVENT_OUT_BUFFER_LOW_QUARTER]=&device->writewq;
            device->event_queue_waiting[EVENT_OUT_BUFFER_HIGH_QUARTER]=&device->writewq;
            device->event_queue_waiting[EVENT_BURST_TRIGGER_READY]=&device->writewq;
            device->event_queue_waiting[EVENT_LOAD_READY]=&device->writewq;
            device->event_queue_waiting[EVENT_END_LOAD_READY]=&device->writewq;
            device->event_queue_waiting[EVENT_DMA_PENDING]=&device->dmawq;

            device->debug_state=0;

            sema_init(&device->sem,1);

            fp->private_data = device;
            // reset DMA engines 

            spin_lock_init(&device->regLock);
            //msg("(%d): resetting DMA engines\n", device->minor);
           // andPLX(device, DMA_CMD_STATUS, STOP_DMA_CMD_0_MASK);
           // andPLX(device, DMA_CMD_STATUS, STOP_DMA_CMD_1_MASK);

            // reset local IRQ flag 
            //msg("(%d): resetting local IRQ flag\n", device->minor);
            DisableIrqLocalAll(device);
            DisableIrqPlx(device);
            readLocal(device,BOARD_CTRL_REG);

            msg("(%d): device_open done\n", device->minor);
            MOD_INC_USE_COUNT;
            return (0);
        }
    }
    msg("(%d): cannot find board\n", MINOR(inode->i_rdev));
    return (-ENODEV);
}

/************************************************************************/
/* close device                                                         */
/************************************************************************/
int device_close(struct inode *inode, struct file *fp)
{
    struct device_board *device = (struct device_board *)fp->private_data;

    msg("(%d): closing board\n", device->minor);

    // reset DMA engines 
    andPLX(device, DMA_CMD_STATUS, STOP_DMA_CMD_0_MASK);

    DisableIrqLocalAll(device);
    DisableIrqPlx(device);

    // free resources
    free_irq(device->irqlevel, device);
    free_pages((u32)device->dma_data[TX].intermediateBuffer, DMA_ORDER);
    
    device->busy = 0;
    MOD_DEC_USE_COUNT;
    return (0);
}

