/***
*** main.c
***
***  General description of this file:
***     Device driver source code for General Standards 16AO12 family of 
***     16-bit analog output boards. This file is part of the Linux
***     driver source distribution for this board.
***     
***  Copyrights (c):
***     General Standards Corporation (GSC),  2002-2006
***
***  Author:
***     Evan Hillman (evan@generalstandards.com)
***
***  Support:
***     Primary support for this driver is provided by GSC. 
***
***  Platform (tested on, may work with others):
***     Linux, kernel version 2.4.x, 2.6.x, Red Hat distribution, Intel hardware.
***/

//////////////////////////////////////////////////////////////////////////
// set the following flag to trace interrupt debug messages.
#ifdef DEBUG
#define TRACE_LOCAL TRUE
#endif

#define __NO_VERSION__

#include "internals.h"

MODULE_LICENSE("GPL");

#define DEFAULT_TIMEOUT MSECS_TO_SLEEP(5000)

static char *regname[] =
{
    "       BOARD_CTRL_REG",                   
    "CHANNEL_SELECTION_REG",              
    "        RATE_CTRL_REG",                     
    "       BUFFER_OPS_REG",                     
    "           RESERVED_1",                         
    "           RESERVED_2",                          
    "       OUTPUT_BUF_REG",                     
    "     ADJUSTABLE_CLOCK",                  
};


/* module load/unload functions */
//int init_module(void);
//void cleanup_module(void);

/* local functions */
//int device_write(struct file *, const char *, size_t, loff_t *);
//int device_read(struct file *, char *, size_t, loff_t *);
//int device_open(struct inode *, struct file *);
//int device_close(struct inode *, struct file *);

//int device_read_buffer(struct device_board *device, char *buf, int nsamp, int noblock);

//int device_read_dma(struct device_board *device, int nsamples);

#ifdef DEBUG
    u32 int_other_count[MAX_BOARDS];
    u32 int_count[MAX_BOARDS];
    u32 dma_count[MAX_BOARDS];
    u32 channel_irq[MAX_BOARDS];
    u32 channel_expected[MAX_BOARDS];
    int board_type[MAX_BOARDS];
    void * context[MAX_BOARDS];
#endif

//#ifdef DEBUG
    /************************************************************************/
    /* regdump - debug only                                                 */
    /************************************************************************/
    void regdump(struct device_board *device,char * label)
    {
        int q;

        for (q=0;q<=LAST_REGISTER;q++)
        {
            errmsg("(%d - %s): register %.2d (addr %.2X) %s value: %.8x\n", device->minor, label, q, q*4, regname[q], readLocal(device,q));
        }

        errmsg("(%d - %s)                 IntCntrlStat: %.8X \n", device->minor, label, readPLX(device,INT_CTRL_STATUS));
        errmsg("(%d - %s)                    DMAArbitr: %.8X \n", device->minor, label, readPLX(device,DMA_MODE_ARBITRATION));
        errmsg("(%d - %s)                 DMAThreshold: %.8X \n", device->minor, label, readPLX(device,DMA_THRESHOLD_REG));
        errmsg("(%d - %s)                   DMA status: %.8X \n", device->minor, label, readPLX(device,DMA_CMD_STATUS));

        errmsg("(%d - %s)                     DMA mode: %.8X \n", device->minor, label, readPLX(device,DMA_CH_0_MODE));
        errmsg("(%d - %s)                  DMAByteCnt0: %.8X \n", device->minor, label, readPLX(device,DMA_CH_0_TRANS_BYTE_CNT));
        errmsg("(%d - %s)                 DMADescrPtr0: %.8X \n", device->minor, label, readPLX(device,DMA_CH_0_DESC_PTR));
        errmsg("(%d - %s)                DMALocalAddr0: %.8X \n", device->minor, label, readPLX(device,DMA_CH_0_LOCAL_ADDR));

        errmsg("(%d - %s)                     DMA mode: %.8X \n", device->minor, label, readPLX(device,DMA_CH_1_MODE));
        errmsg("(%d - %s)                  DMAByteCnt0: %.8X \n", device->minor, label, readPLX(device,DMA_CH_1_TRANS_BYTE_CNT));
        errmsg("(%d - %s)                 DMADescrPtr0: %.8X \n", device->minor, label, readPLX(device,DMA_CH_1_DESC_PTR));
        errmsg("(%d - %s)                DMALocalAddr0: %.8X \n", device->minor, label, readPLX(device,DMA_CH_1_LOCAL_ADDR));

        errmsg("Intermediate buffer data:\n");
        for (q=0;q<=12;q++)
        {
            printk("%.8X ",device->dma_data[TX].intermediateBuffer[q]);
        }
        errmsg("\n----------------------------------\n");
    }
//#endif

/************************************************************************/
/* timeout_handler                                                      */
/************************************************************************/
void timeout_handler(unsigned long ptr)
{
    struct device_board *device = (struct device_board *) ptr;

    int i;

    mark(0xAaaaaaaa);
    regdump(device,"timeout " );
    device ->timeout=TRUE;

    // wake up whatever was waiting.
    errmsg(" timeout -> plx int %.8X board control %.8X DMA status %.8X \n",
        readPLX(device, INT_CTRL_STATUS), readLocal(device,BOARD_CTRL_REG), readPLX(device, DMA_CMD_STATUS));

    // wake up whatever was waiting before the timeout
    // wake up whatever was waiting.
    errmsg(" timeout -> plx int %.8X board control %.8X DMA status %.8X \n",
        readPLX(device, INT_CTRL_STATUS), readLocal(device,BOARD_CTRL_REG), readPLX(device, DMA_CMD_STATUS));
    errmsg(" DMA Mode %.8X\n",readPLX(device, DMA_CH_0_MODE));
    errmsg(" Firm rev %.8X DMA event # %d\n",readLocal(device,0x3c/4),EVENT_DMA_PENDING);

    for(i=0;i<=EVENT_LAST_IRQ_EVENT;i++)
    {
        //errmsg("(%d): ERROR - event %d status %d\n", device->minor, i,atomic_read(&device->irq_event_pending[i]));
        if (atomic_read(&device->irq_event_pending[i]))
        {
            errmsg("(%d): ERROR - event %d timeout\n", device->minor, i);
            clearEvent(i);
            wake_up(device->event_queue_waiting[i]);
        }
    }

    errmsg("(%d): End timeout handler\n", device->minor);
}

/*************************************************/
#define CLEAR_SPURIOUS_PLX(regval,mask, desc)  { \
    if (regval & mask) \
{ \
    writePLX(device, INT_CTRL_STATUS, (regval & ~mask)); \
} \
}
/*errmsg("(%d): Unexpected interrupt %s %.8lX %.8X %.8X\n",device->minor,desc, regval,mask,~mask); \*/

/************************************************************************/
/* Interrupt handler                                                    */
/************************************************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
irqreturn_t device_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#else
void device_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
    struct device_board *device = (struct device_board *)dev_id;
    unsigned long plxreg;
    unsigned long localreg,bcr_reg;
    int handled=TRUE;
    int local_irq;

    msg(": Interrupt()\n");
    mark(0xbee00000);

    // determine if the interrupt is from this board.

    plxreg = readPLX(device,INT_CTRL_STATUS);
    if (((plxreg & IRQ_PCI_ENABLE) ==0) ||
        (((plxreg & IRQ_DMA_0_ACTIVE)==0)&&((plxreg & IRQ_LOCAL_ACTIVE))==0)) 
    { // not ours
        msg("Interrupt not ours...\n");
#ifdef DEBUG
        int_other_count[device->board_index]++;
#endif
        mark(0x11111111);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        return IRQ_NONE;
#else
        return;             
#endif
    }

    bcr_reg = readLocal(device,BOARD_CTRL_REG);
    msg("(%d): Interrupt!, %.8lX\n", device->minor,bcr_reg);
    mark(0x12121212);
    // deal with local interrupt sources.

    DisableIrqPlx(device);
    andLocal(device,BOARD_CTRL_REG, (~BCR_INTERRUPT_REQUEST_FLAG));

    // check for DMA complete interrupt 
    if (atomic_read(&device->irq_event_pending[EVENT_DMA_PENDING])) {
        //msg("(%d): IRQ Dma in progress...\n",device->minor);
        localreg = readPLX(device, DMA_CMD_STATUS);
        if (localreg & DMA0_DONE) {
            //msg("(%d): IRQ Dma done...\n",device->minor);
            // acq interrupt 
            orPLX(device, DMA_CMD_STATUS, DMA0_CLR_INT);
            orPLX(device, DMA_CMD_STATUS,(DMA0_CLR_INT|DMA1_CLR_INT)); 
            // wake up process blocked in 'read()' 
            clearEvent(EVENT_DMA_PENDING);
#ifdef DEBUG
            dma_count[device->board_index]++;
#endif
            wake_up(&device->dmawq);
        }
    }
    else    // check for and clear spurious DMA interrupts...
    {
        plxreg = readPLX(device, DMA_CMD_STATUS);
        if (plxreg & DMA0_ENABLE) {
            // wipe out the DMA1 bits while here, better safe than sorry.
            msg("(%d): Unexpected interrupt DMA0_ENABLE\n",device->minor); 
            writePLX(device, DMA_CMD_STATUS, DMA0_CLR_INT|DMA1_CLR_INT); 
        }
    }

    // check for local interrupt 
    DisableIrqLocalAll(device);

    // check for local interrupt

    mark(0x13131313);

    if (bcr_reg & BCR_INTERRUPT_REQUEST_FLAG) {
        handled=FALSE;

        // see which kind of event is this.  The hardware is assumed to only have one
        // interrupt active at a time.
#ifdef TRACE_LOCAL
        int_count[device->board_index]++;
#endif
        local_irq=((bcr_reg&BCR_IRQ_MASK)>>BCR_IRQ_SHIFT); 

        msg("(%d): local IRQ # %d active: %d\n", device->minor,local_irq, atomic_read(&device->irq_event_pending[local_irq]));
        if (atomic_read(&device->irq_event_pending[local_irq])) {
            mark(0x14141414);
            msg("event %d active\n",local_irq);
            clearEvent(local_irq);
            wake_up(device->event_queue_waiting[local_irq]);
            handled = TRUE;
        }
        else
        {
            mark(0x15151515);
            errmsg("(%d): disabled irq on device %p\n", device->minor, device);
        }
    }

    if (!handled)
    {
#ifdef TRACE_LOCAL
        mark(0x16161616);
        plxreg = readPLX(device, INT_CTRL_STATUS);
        bcr_reg = readLocal(device,BOARD_CTRL_REG);
#endif
        errmsg("(%d): irq handler - interrupt not handled!\n", device->minor);
    }
#ifdef DEBUG
    msg("(%d): irq handler re-enable %p\n", device->minor,device);
#endif

#ifdef DEBUG
    //bcr_reg = readLocal(device,BOARD_CTRL_REG);
    //msg("(%d): irq handler re-enable PLX: %.8lx BCR: %.8lx\n", device->minor,plxreg,bcr_reg);
#endif

    /*
    *   check every possible interrupt source, to ensure that the
    * current interrupt is turned off, expected or not.
    */

    // PLX interrupt-control registers.
    CLEAR_SPURIOUS_PLX(plxreg,IRQ_LOCAL_LSERR_ABORT, "IRQ_LOCAL_LSERR_ABORT*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_LOCAL_LSERR_ABORT, "IRQ_LOCAL_LSERR_ABORT*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_LOCAL_LSERR_OVERFLOW, "IRQ_LOCAL_LSERR_OVERFLOW*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_GENERATE_SERR, "IRQ_GENERATE_SERR*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_MAILBOX_ENABLE, "IRQ_MAILBOX_ENABLE*");
    //  CLEAR_SPURIOUS_PLX(plxreg, IRQ_PCI_ENABLE, "IRQ_PCI_ENABLE*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_PCI_DOORBELL_ENABLE, "IRQ_PCI_DOORBELL_ENABLE*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_ABORT_ENABLE, "IRQ_ABORT_ENABLE*");
    //  CLEAR_SPURIOUS_PLX(plxreg, IRQ_LOCAL_PCI_ENABLE, "IRQ_LOCAL_PCI_ENABLE*");
    //  CLEAR_SPURIOUS_PLX(plxreg, IRQ_LOCAL_ENABLE, "IRQ_LOCAL_ENABLE*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_LOCAL_DOORBELL_ENABLE, "IRQ_LOCAL_DOORBELL_ENABLE*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_DMA_1_ENABLE,"IRQ_DMA_1_ENABLE*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_LOCAL_DOORBELL_ACTIVE,"IRQ_LOCAL_DOORBELL_ACTIVE*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_DMA_1_ACTIVE,"IRQ_DMA_1_ACTIVE*");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    return IRQ_HANDLED;
#else
    return;             
#endif

}
